/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


//This program generates codesets given the codeset paramaters. 
//The codeset is written to a file called "codes.N.A.w.K.txt"
//where (N, A, w, K) are the codeset parameters. 
//The format of the codeset file is described in the simone design document

//Note that the codeset generates codewords with K=w. (The K passed as a command 
//line argument is NOT used.


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

int N, w, A, K;
int * codeset;
int numcodewords;
FILE *fout;

void gen1D();
void gen2D();

void DumpCodeword(int index);



int main(int argc,  char **argv)
{

	
	if(argc != 7)
	{
		printf("Usage: codegen <seed> <N> <A> <w> <K> <numcodewords>\n");
		exit(0);
	}

	N = atoi(argv[2]);
	A = atoi(argv[3]);
	w = atoi(argv[4]);
	K = atoi(argv[5]);

	int seed = atoi(argv[1]);
	srandom(seed);
	
	char codefile[20];
	sprintf(codefile, "codes.%d.%d.%d.%d.txt", N, A, w, K);
	
	fout = fopen(codefile, "w");
	if (fout == NULL)
	{
		printf("Could not open file for writing\n");
		exit(-1);
	}

	numcodewords =atoi(argv[6]);
	
	fprintf(fout, "%d\n", numcodewords);
	fprintf(fout, "%d\n", w);

        codeset = new int [numcodewords * A * N];
        for(int count = 0; count < numcodewords *A *N; count++)
		codeset[count] = 0;

	if (A == 1)
		gen1D();
	else
		gen2D();

}

void gen1D()
{

	for(int count =0; count < numcodewords; count++)
	{

		//printf("Generating 1D\n");	
		int wt = 0;
		while(true)
		{
			int c = (int) ((random()/(pow(2, 31) - 1))*N);
			if (codeset[count * (A*N) +  c] == 1)
				continue;
		
			codeset[count * (A*N) +  c] = 1;
				 		
			wt++;
			if(wt == w)
				break;

		}
		for (int count3 =N-1; count3 >=0; count3--)	
		{
			fprintf(fout,"%d ", codeset[count* A* N + count3]);	
		}
		fprintf(fout, "\n");	

		//printf("Code:\n");
		DumpCodeword(count);
		printf("\n");
	}


}	










void gen2D()
{
	int * chosenw = new int[A];

//One 1 per wavelength	, w wavelengths overall
	for(int count =0; count < numcodewords; count++)
	{
	
		//printf("Generating 2D\n");	
		for(int count2 =0; count2 <A;count2++)
			chosenw[count2] = 0;	
		for (int count2 = 0; count2 < w; count2++)
		{
			while(true)
			{
				int wavelength = (int)((random()/(pow(2, 31) - 1))*A);
				if(chosenw[wavelength] == 1)
					continue;
					
				//printf("Chosen: %d\n", wavelength);	
				chosenw[wavelength] = 1;
				break;
			}					
		}


		for (int count2 = 0; count2 < A; count2++)
		{
			int codeword = 0;

			if(chosenw[count2] == 1)
			{
				int c = (int) ((random()/(pow(2, 31) - 1))*N);
				codeset[count * (A*N) + count2 * N + c] = 1;
			}
		
			for (int count3 =N-1; count3 >=0; count3--)	
			{
				fprintf(fout,"%d ", codeset[count* A* N + count2 * N +  count3]);	
			}
	
			fprintf(fout, "\n");	
		}
		fprintf(fout, "\n");	

		//printf("Code:\n");
		DumpCodeword(count);
		printf("\n");


	}	

	exit(0);
}

void DumpCodeword(int index)
{
	for (int count2 =0; count2 < A; count2++)	
	{	
		for (int count3 =0; count3 < N; count3++)	
		{

			printf("%d ", codeset[index * (A*N) + count2 * N + count3]);
		}
		printf("\n");
	}
}

