/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

//This program converts a codeset file from the GECO format to the format 
//understood by simone.
//Refer to the GECO documentation for information about the GECO format
//Refer to the simone design document for information about the simone format.


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

int N, w, A, K;
int * codeset;

void DumpCodeword(int index);
int main(int argc,  char **argv)
{

	
	if(argc != 5)
	{
		printf("Usage: geco2simone  <N> <A> <w> <K>\n");
		exit(0);
	}

	N = atoi(argv[1]);
	A = atoi(argv[2]);
	w = atoi(argv[3]);
	K = atoi(argv[4]);

	char codefile[20];
	sprintf(codefile, "codes.%d.%d.%d.%d.geco.txt", N, A, w, K);
	
	FILE *fout = fopen(codefile, "r");
	if (fout == NULL)
	{
		printf("Could not open GECO file for reading\n");
		exit(-1);
	
	}
	int numchips = 0;
	int chip;
	while(fscanf(fout, "%d", &chip) != EOF)
		numchips++;
	rewind(fout);
	
	int numcodewords =numchips/w;

	printf("Number of codewords: %d\n", numcodewords);
	codeset = new int[numcodewords * A *N];
	for(int count = 0; count < numcodewords * A *N ; count++)
		codeset[count] = 0;
 
	for(int count = 0; count < numcodewords; count++)
	{
		for(int count2 =0; count2 < w; count2++)
		{
			fscanf(fout, "%d", &chip);
			codeset[count*(A*N) + chip] = 1;
		}
		//DumpCodeword(count);
		//printf("\n");
	}

	fclose(fout);
	sprintf(codefile, "codes.%d.%d.%d.%d.txt", N, A, w, K);
	fout = fopen(codefile, "w");
	if (fout == NULL)
	{
		printf("Could not open file for writing\n");
		exit(-1);
	}
	fprintf(fout, "%d\n", numcodewords);
	fprintf(fout, "%d\n", w);


	for(int count =0; count < numcodewords; count++)
	{
		for(int count2 =0; count2 <A;count2++)
		{
			for (int count3 = 0; count3 < N; count3++)
			{
				fprintf(fout, "%d ", codeset[count *(A*N) + count2*N + count3]);
			}
		}
		fprintf(fout, "\n");
	}

	fclose(fout);


	exit(0);
}

void DumpCodeword(int index)
{
	
	for (int count2 =0; count2 < A; count2++)	
	{	
		for (int count3 =0; count3 < N; count3++)	
		{
			printf("%d ", codeset[index * (A*N) + count2 * N + count3]);
		}
		printf("\n");
	}
}

