/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <defs.h>
#include <packet.h>
#include <eventhandler.h>
#include <squeue.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <receiver.h>
#include <trdump.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>


extern Scheduler * sched;
extern InTransit * it;
extern char outputfile[20];
extern Codeset *codeset;

extern int propSpeed;


//Implemets the functions of a receiver. 
//Give a packet and the corresponding state at a point in time/sapce, it calculates whether that packet will be received correctly
//Currently only the correlation receiver is implemented.
 


Receiver::Receiver():EventHandler()
{
	debug = 0;
}

Receiver::~Receiver()
{
}


void Receiver::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend


}
//The event handling function of the Receiver. It dequeues a packet and
//sends it to the next component.
void Receiver::handleevent(Packet *p)
{
	//dtprintf("receiver.cc: Receiver::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	//dtprintf("receiver.cc: Receiver::handleevent(): Packet id: %d\n", p->id);
	//dtprintf("receiver.cc: Receiver::handleevent(): Packet processing complete, passing it on to the next component.\n");


}




void Receiver::dumpstats()
{
	printf("receiver.cc: Receiver::dumpstats():\n");
}


CorrelationReceiver::~CorrelationReceiver()
{
}

CorrelationReceiver::CorrelationReceiver(int n, int a, int W, int K): Receiver()
{
	A = a;	
	N = n;
	w = W;
}

int Receiver::receive(Section * s, Packet *p)
{

	return 0;
}


int CorrelationReceiver::receive(Section * s, Packet *rp)
{


	double x = s->head;

	dtprintf("receive.cc: Receive::receive(): Calculating state\n"); 		
	dtprintf("receive.cc: Receive::receive(): Calculating state at : %.6f, %.6f, %.6f\n", x, 0, 0); 		
	int * rpstate = rp->calcstate(x, 0, 0, 0, 0); 

	//Note it may be shifted because it is relative to the section head.
	dtprintf("receive.cc: Receive::receive(): Codeword: \n"); 		
	for (int count = 0; count < (rp->A)*(rp->N); count++)
		dprintf("%d ", rpstate[count]); 		
	dprintf("\n"); 		


	int * state = new int [rp->A *rp->N];
	for (int count = 0; count < (rp->A)*(rp->N); count++)
		state[count] =  0;

	int pcount =0; 
	Packet*p = s->q->head;
	int overlapcount[w];
	for (int count = 0; count < w; count++)
		overlapcount[count] = 0;


	while(p!= NULL)
		{
		dtprintf("receive.cc: Receive::receive(): Checking packet: %d\n", p->id); 		
		int * pstate = p->calcstate(x, 0, 0 , 0, 0);
		int wt = 0;				
		for (int count = 0; count < rp->A*rp->N; count++)
		{
			state[count]+= pstate[count];

			if((pstate[count] == 1) && (rpstate[count] == 1))
			{
				//Count the number of overlapping chips in each 1 chip postion
				//(Assume they come from different codewords)
				overlapcount[wt]++;
			}	
			if(pstate[count] == 1)
				wt++;

		}	
		delete pstate;	
		pcount++;
		p = p->next;
	}

	dtprintf("receive.cc: Receive::receive(): Numonline: %d\n", pcount); 		
	dtprintf("receive.cc: Receive::receive(): State: \n"); 		
	for (int count = 0; count < (rp->A)*(rp->N); count++)
		dprintf("%d ", state[count]); 		
	dprintf("\n"); 		
		


	dtprintf("receive.cc: Receive::receive(): Checked all packets\n"); 		
//Simple receiver
/*	int overlaps = 0;	
	for (int count = 0; count < rp->A*rp->N; count++)
		if ((rpstate[count] == 1) && (state[count] > 1))
			overlaps++;

	dtprintf("receive.cc: Receive::receive(): Overlaps: %d\n", overlaps); 		

	if(overlaps == rp->w)
	{
		rp->status = PACKET_ERROR_MUI;
	}
*/

//Real performance
	int overlaps = 0;	
	for (int count = 0; count < rp->A*rp->N; count++)
		if ((rpstate[count] == 1) && (state[count] > 1))
			overlaps++;

	dtprintf("receive.cc: Receive::receive(): Overlaps: %d\n", overlaps); 		

	if(overlaps == rp->w)
	{

		//Pe = Probability received codeword is a zero bit and all 
		//overlapping codeswords are ones
		double pe = 1;	
		//At least one of the codewords in all of the overlapping chips must be on
		for(int count =0; count < w; count++)
		{
			
			dtprintf("receive.cc: Receive::receive(): Overlapping packets: %d\n", overlapcount[count]); 		
			pe = pe * (1-pow( (1- 0.5), overlapcount[count]));
		}	

		//And this chip must be on
		pe = 0.5 * pe;
		dtprintf("receive.cc: Receive::receive(): Pr bit error: %.6f\n", pe); 		

		//There are as many trials as the number of bytes in the section
		double length = s->tail - s->head;
		dtprintf("receive.cc: Receive::receive(): Length: %.6f\n", length); 		
		double chiptime = (1.0/codeset->chiprate);
		double chiplength = chiptime * propSpeed;
		double numchips = length/chiplength;

		dtprintf("receive.cc: Receive::receive(): Numchips: %.6f\n", numchips); 		
		int bytes = (int)(numchips/(double)N);
		dtprintf("receive.cc: Receive::receive(): Bytes: %d\n", bytes); 	

		//This must happen at least once in the section	
		double proberror = 1 -  pow((1- pe), bytes);	
		dtprintf("receive.cc: Receive::receive(): Pe: %.6f\n", proberror); 		
		
		double rnd = (random()/(pow(2, 31) -1));
		if(rnd <= proberror)
			rp->status = PACKET_ERROR_MUI;

	}
	delete state;
	delete rpstate;
	return 0;
}



void CorrelationReceiver::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend


}
//The event handling function of the CorrelationReceiver. It dequeues a packet and
//sends it to the next component.
void CorrelationReceiver::handleevent(Packet *p)
{
	//dtprintf("receiver.cc: CorrelationReceiver::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	//dtprintf("receiver.cc: CorrelationReceiver::handleevent(): Packet id: %d\n", p->id);
	//dtprintf("receiver.cc: CorrelationReceiver::handleevent(): Packet processing complete, passing it on to the next component.\n");


}

void CorrelationReceiver::dumpstats()
{
	printf("receiver.cc: CorrelationReceiver::dumpstats():\n");
}
