/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <defs.h>
#include <packet.h>
#include <eventhandler.h>
#include <squeue.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <txsched.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <receiver.h>
#include <stateest.h>
#include <state.h>


extern Scheduler * sched;
extern InTransit * it;
extern TrDump * trdump;
extern TrGen * trgen;
extern char outputfile[20];
extern Codeset *codeset;
extern Receiver *rcvr;
extern StateEstimator *stateest;

extern int propSpeed;

TxSched::TxSched():EventHandler()
{
	debug = 0;
}

TxSched::~TxSched()
{
}


void TxSched::process(Packet *p)
{


}


//The event handling function of the TxSched. It dequeues a packet and
//sends it to the next component.
void TxSched::handleevent(Packet *p)
{
	dtprintf("txsched.cc: TxSched::handleevent(): Entered function\n");
}



void TxSched::dumpstats()
{

}



TxSched_Aloha::TxSched_Aloha():TxSched()
{
}

TxSched_Aloha::~TxSched_Aloha()
{
}


void TxSched_Aloha::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_Aloha::process(): Scheduling packet: %d\n", p->id); 		

//Stats collection
	//If no stats are being collected right now
	if(stateest->reading == 0)
	{

		int * state = stateest->estimatestate(p->depx, 0, 0, p->depx, 0, 0);
		stateest->oldstate = state;
		stateest->estx = p->depx;
		//Schedule stats collections
		sched->scheduleevent(stateest, NULL, stateest->delay);
	}


	sched->scheduleevent(this, p, 0);	

}


//The event handling function of the TxSched_Aloha. It dequeues a packet and
//sends it to the next component.
void TxSched_Aloha::handleevent(Packet *p)
{
	dtprintf("txsched.cc: TxSched_Aloha::handleevent(): Entered function\n");
	dprintf("\n");
	p->departuretime = sched->currenttime; 	 
	p->offset =0;	
	it->add(p);
	trdump->process(p);
}



void TxSched_Aloha::dumpstats()
{

}


TxSched_Selfish::TxSched_Selfish(int n, int a, int W, int k):TxSched()
{

	N = n;
	A = a;
	w = W;
	K= k;
}

TxSched_Selfish::~TxSched_Selfish()
{
}


void TxSched_Selfish::process(Packet *p)
{
	dtprintf("txsched.cc: TxSched_Selfish::process(): Scheduling packet: %d\n", p->id);
	dtprintf("txsched.cc: TxSched_Selfish::process(): Codeword :\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}
	dprintf("\n");

	int * state;

	

//Perfect state estimation
/*	
	int * tstate = stateest->estimateState(p->depx, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_Selfish::process(): True state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			dprintf("%d ", tstate[count1*N + count]);
		}
		dprintf("\n");
	
	}
*/



//Realistic state estimation: Assume meeting point is at (0,0,0), zero delay between 
//meeting point and nodes
	state = stateest->estimatestate(0, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_Selfish::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N -1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}



//Stats collection
/*
	double totalerrorsq= 0;
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			double errorsq = pow (tstate[count1*N + count] - state[count1*N + count], 2);	
			totalerrorsq+=errorsq;	
		}
	}
	double error = pow(totalerrorsq, 0.5);
	dprintf("Error: %.6f\n", error);
*/




//Stats collection
	//If no stats are being collected right now
	/*if(stateest->reading == 0)
	{
		//Start stats collection
		stateest->oldstate = state;
		stateest->estx = p->depx;
		//Schedule stats collections
		sched->scheduleevent(stateest, NULL, stateest->delay);
	}
	*/


	int offsetcount =0;
	int offset[w * N];

	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =N-1; count >=0; count--)
		{
			if (state[N*count1 + count] == 0)
			{
				for (int count2 = N-1; count2 >=0; count2--)
				{
					if(codeset->set[(p->codeword) * (A*N) + N*count1 + count2] ==1)
					{
						//dtprintf("txsched.cc: TxSched_Selfish::process(): Wavelength: %d State zero: %d Code one: %d\n", count1, count, count2 );
						if (count2 >= count)
							offset[offsetcount] = count2 - count;
						else
							offset[offsetcount] = N - (count - count2); 
						dtprintf("txsched.cc: TxSched_Selfish::process(): Found a possible offset: %d\n", offset[offsetcount]);
						offsetcount++;
					}	
					
				}

			}
		}		
	}


	if (offsetcount != 0)
	{

		int rnd = (int)(floor((random()/(pow(2, 31) -1)) *offsetcount));
		int off = offset[rnd];

		double timeoffset = sched->currenttime;
		double chiptime = 1.0/codeset->chiprate;
		int chipoffset = int((timeoffset/chiptime) + 0.5);
		chipoffset = chipoffset % (int) N;	
		int normoffset  = (chipoffset + off) % N;
	
		dtprintf("txsched.cc: TxSched_Selfish::handle_event:: Sending at offset (relative to estimation time): %d\n", off);
		dtprintf("txsched.cc: TxSched_Selfish::handle_event:: Sending at offset (relative to normalized time): %d\n", normoffset);
		


		double delay = (double)off * chiptime; 	
		p->departuretime = sched->currenttime + delay;

		p->calcnormalizedstate();

		p->offset = off;
		it->add(p);
		dtprintf("txsched.cc: TxSched_Selfish::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n", p->id);
	}
	

//TBR
	//delete state;
	dprintf("\n");
}


//The event handling function of the TxSched_Selfish. It dequeues a packet and
//sends it to the next component.
void TxSched_Selfish::handleevent(Packet *p)
{

		dtprintf("txsched.cc: TxSched_Selfish::handleevent(): Entered function\n");
		dprintf("\n");
		trdump->process(p);
}



void TxSched_Selfish::dumpstats()
{

}


TxSched_Threshold::TxSched_Threshold(int n, int a, int W, int k, double threshold):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = threshold* N *A;
}

TxSched_Threshold::~TxSched_Threshold()
{
}


void TxSched_Threshold::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_Threshold::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_Threshold::process(): Threshold: %.6f\n", th); 		
	
	dtprintf("txsched.cc: TxSched_Threshold::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}


//TBD: Change point of state estimation
	//int * state = stateest->EstimateState(p->depx, 0, 0, p->depx, 0, 0);
	int * state = stateest->estimatestate(0, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_Threshold::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}


	int offset = -1;	
	int overlaps = 0;

		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =N-1; count >=0; count--)
			{
				if (state[N*count1 + count] == 0)
				{
					for (int count2 = N-1; count2 >=0; count2--)
					{
						if(codeset->set[(p->codeword) * (A*N) + N*count1 + count2] ==1)
						{

							dtprintf("txsched.cc: TxSched::handle_event::SC: %d CC: %d\n", count, count2 );
							if (count2 >= count)
								offset = count2 - count;
							else
								offset = N - (count - count2); 
							int * newcodeword = codeset->RightShiftCodeword(p->codeword, offset);	
							for (int count3 =0; count3 < A; count3++)
							{
								for (int count4=0; count4< N; count4++)
								{
									if (state[N*count3 + count4] + newcodeword[N*count3+count4] > 1) 
										overlaps++; 		
								}
							}
							if(overlaps < th)
							{
								delete newcodeword;
								break;
							}
							else
							{
								offset = -1;
								delete newcodeword;
							}
						}	
					
					}
				if(offset != -1)
					break;

			}
			if(offset != -1)
				break;
		}		
		if(offset != -1)
			break;
	}
	if (offset != -1)
	{
		dtprintf("txsched.cc: TxSched::handle_event::Choosing offset: %d\n", offset );
		double chiptime = (1.0/p->chiprate);
		double delay = (double)offset * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = offset;
		it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n");
	}

	delete state;
	dprintf("\n");



}


//The event handling function of the TxSched_Threshold. It dequeues a packet and
//sends it to the next component.
void TxSched_Threshold::handleevent(Packet *p)
{
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_Threshold::handleevent(): Entered function\n");
}



void TxSched_Threshold::dumpstats()
{

}

TxSched_OverlapSection::TxSched_OverlapSection(int n, int a, int W, int k, double threshold):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = threshold* N;
}

TxSched_OverlapSection::~TxSched_OverlapSection()
{
}


void TxSched_OverlapSection::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_OverlapSection::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_OverlapSection::process(): OverlapSection: %.6f\n", th); 		

	dtprintf("txsched.cc: TxSched_OverlapSection::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}


//TBD: Change point of state estimation
	//int * state = stateest->EstimateState(p->depx, 0, 0, p->depx, 0, 0);
	int * state = stateest->estimatestate(0, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_OverlapSection::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}

	int offset = -1;	
	int overlaps = 0;
	int ones = 0;

		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =N-1; count >=0; count--)
			{
				if (state[N*count1 + count] == 0)
				{
					for (int count2 = N-1; count2 >=0; count2--)
					{
						if(codeset->set[(p->codeword) * (A*N) + N*count1 + count2] ==1)
						{

							dtprintf("txsched.cc: TxSched::handle_event::SC: %d CC: %d\n", count, count2 );
							if (count2 >= count)
								offset = count2 - count;
							else
								offset = N - (count - count2); 
							int * newcodeword = codeset->RightShiftCodeword(p->codeword, offset);	
							for (int count3 =0; count3 < A; count3++)
							{
								for (int count4=0; count4< N; count4++)
								{
									if (state[N*count3 + count4] + newcodeword[N*count3+count4] > 1) 
										overlaps ++; 		
									if (state[N*count3 + count4] + newcodeword[N*count3+count4] == 1) 
										ones ++; 		
								}
							}
							if(overlaps < ones)
							{
								delete newcodeword;
								break;
							}
							else
							{
								delete newcodeword;
								overlaps = 0;
								offset = -1;
							}
						}	
					
					}
				if(offset != -1)
					break;

			}
			if(offset != -1)
				break;
		}		
		if(offset != -1)
			break;
	}
	if (offset != -1)
	{
		dtprintf("txsched.cc: TxSched::handle_event::Choosing offset: %d\n", offset );
		double chiptime = (1.0/p->chiprate);
		double delay = (double)offset * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = offset;
		it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n");
	}


	delete state;
	dprintf("\n");


}


//The event handling function of the TxSched_OverlapSection. It dequeues a packet and
//sends it to the next component.
void TxSched_OverlapSection::handleevent(Packet *p)
{
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_OverlapSection::handleevent(): Entered function\n");
}



void TxSched_OverlapSection::dumpstats()
{

}

TxSched_OverlapSize::TxSched_OverlapSize(int n, int a, int W, int k, double threshold):TxSched()
{

	N = n;
	A =a;
	w = W;
	K=k;
	th = threshold;
}

TxSched_OverlapSize::~TxSched_OverlapSize()
{
}


void TxSched_OverlapSize::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend
	dtprintf("trdump.cc: TxSched_OverlapSize::process(): Scheduling a packet\n"); 		
	dtprintf("trdump.cc: TxSched_OverlapSize::process(): OverlapSize: %.6f\n", th); 		
	dtprintf("txsched.cc: TxSched_OverlapSize::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}


//TBD: Change point of state estimation
	//int * state = stateest->EstimateState(p->depx, 0, 0, p->depx, 0, 0);
	int * state = stateest->estimatestate(0, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_OverlapSize::process(): Estimated state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}



	int offset = -1;	
	int donotsend = 0;

		for (int count1 =0; count1 < A; count1++)
		{
			for (int count =0; count < N; count++)
			{
				if (state[N*count1 + count] == 0)
				{
					for (int count2 = 0; count2 < N; count2++)
					{
						if(codeset->set[(p->codeword) * (A*N) + N*count1 + count2] ==1)
						{

							dtprintf("txsched.cc: TxSched::handle_event::SC: %d CC: %d\n", count, count2 );
							if (count >= count2)
								offset = count - count2;
							else
								offset = N - (count2 - count); 
							int * newcodeword = codeset->RightShiftCodeword(p->codeword, offset);	
							for (int count3 =0; count3 < A; count3++)
							{
								for (int count4=0; count4< N; count4++)
								{
									if (state[N*count3 + count4] + newcodeword[N*count3+count4] > 1) 
										donotsend = 1; 		
								}
							}
							if(donotsend == 1)
							{
								delete newcodeword;
								donotsend = 0;
								offset = -1;
							}
							else
							{
								delete newcodeword;
								break;
							}
						}	
					
					}
				if(offset != -1)
					break;

			}
			if(offset != -1)
				break;
		}		
		if(offset != -1)
			break;
	}
	if (offset != -1)
	{
		dtprintf("txsched.cc: TxSched::handle_event::Choosing offset: %d\n", offset );
		double chiptime = (1.0/p->chiprate);
		double delay = (double)offset * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = offset;
		it->add(p);
		dtprintf("txsched.cc: TxSched::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n");
	}



	delete state;
	dprintf("\n");

}


//The event handling function of the TxSched_OverlapSize. It dequeues a packet and
//sends it to the next component.
void TxSched_OverlapSize::handleevent(Packet *p)
{
	trdump->process(p);
	dtprintf("txsched.cc: TxSched_OverlapSize::handleevent(): Entered function\n");
}



void TxSched_OverlapSize::dumpstats()
{

}

TxSched_OptimalState::TxSched_OptimalState(int n, int a, int W, int k):TxSched()
{

	N = n;
	A = a;
	w = W;
	K= k;
}

TxSched_OptimalState::~TxSched_OptimalState()
{
}


void TxSched_OptimalState::process(Packet *p)
{
	dtprintf("txsched.cc: TxSched_OptimalState::process(): Scheduling packet: %d\n", p->id);
	dtprintf("txsched.cc: TxSched_OptimalState::process(): Code:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", codeset->set[p->codeword *(A*N) + count1*N + count]);
		}
		dprintf("\n");
	}

	int * state;

	

//TBD: Change point of state estimation
//Perfect state estimation
/*	int * tstate = stateest->estimatestate(p->depx, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_OptimalState::process(): True state:\n");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", tstate[count1*N + count]);
		}
		dprintf("\n");
	}
*/
//Realistic state estimation: Assume meeting point is at (0,0,0), zero delay between meeting point and 
//nodes
	state = stateest->estimatestate(0, 0, 0, p->depx, 0, 0);
	dtprintf("txsched.cc: TxSched_OptimalState::process(): Estimated state:\n");
	dtprintf("State:");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}
/*
	double totalerrorsq= 0;
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			double errorsq = pow (tstate[count1*N + count] - state[count1*N + count], 2);	
			totalerrorsq+=errorsq;	
		}
	}
	double error = pow(totalerrorsq, 0.5);
	dprintf("Error: %.6f\n", error);
*/




//Stats collection
	//If no stats are being collected right now
	/*if(stateest->reading == 0)
	{

		stateest->oldstate = state;
		stateest->estx = p->depx;
		//Schedule stats collections
		sched->scheduleevent(stateest, NULL, stateest->delay);
	}
	*/

	int offset = -1;	
	int minoffset = -1;	
	double minoptstdist = -1;	
	double n0=0;
	double n1 =0;	
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			if (state[N*count1 + count] == 0)
			{
				for (int count2 = 0; count2 < N; count2++)
				{
					if(codeset->set[(p->codeword) * (A*N) + N*count1 + count2] ==1)
					{
						printf("txsched.cc: TxSched_OptimalState::process(): Wavelength: %d State zero: %d Code one: %d\n", count1, count, count2 );
						if (count >= count2)
							offset = count - count2;
						else
							offset = N - (count2 - count); 

						int * newc = codeset->RightShiftCodeword(p->codeword, offset);			
	
						for (int count3 =0; count3 < A; count3++)
						{
							for (int count4 =0; count4 < N; count4++)
							{
								int news = state[N*count1+ count] + newc[N*count3 + count4];
								if(news == 0)
									n0++;
								else if(news == 1)
									n1++;
							}
						}
						//TBD: Change distance calc									
						double optstdist = pow((pow(n0, 2) + pow((n1 - (N-(w-1))), 2)), 0.5); 
						if ((optstdist < minoptstdist) || (minoptstdist == -1))
						{
							printf("txsched.cc: TxSched_OptimalState::process(): iFound lower distance\n");
							minoptstdist = optstdist;
							minoffset = offset;
						}					
			
					}	
					
				}

			}
		}		
	}
	if (minoffset != -1)
	{
		printf("txsched.cc: TxSched_OptimalState::handle_event::Choosing offset: %d\n", offset );
		double chiptime = (1.0/p->chiprate);
		double delay = (double)minoffset * chiptime; 	
		p->departuretime = sched->currenttime + delay;
		p->offset = minoffset;
		it->add(p);
		dtprintf("txsched.cc: TxSched_OptimalState::handle_event::Departure time: %.12f\n", p->departuretime );
		sched->scheduleevent(this, p, delay);	
	}
	else
	{
		dtprintf("txsched.cc: TxSched::handle_event::Cannot schedule packet: %d\n", p->id);
	}
	

//TBR
	//delete state;
	dprintf("\n");
}


//The event handling function of the TxSched_OptimalState. It dequeues a packet and
//sends it to the next component.
void TxSched_OptimalState::handleevent(Packet *p)
{
	dtprintf("txsched.cc: TxSched_OptimalState::handleevent(): Entered function\n");
	dprintf("\n");
	trdump->process(p);
}



void TxSched_OptimalState::dumpstats()
{

}


		
